package ca.recrown.islandsurvivalcraft.caching;

import java.util.HashMap;
import java.util.LinkedList;

public class Cache<Key, Value> {
    private final int maxSize;
    private final HashMap<Key, Value> data;
    private final HashMap<Key, Integer> occurrences;
    private final LinkedList<Key> occurrenceOrder;

    public Cache(int maxSize) {
        data = new HashMap<>(maxSize);
        occurrenceOrder = new LinkedList<>();
        this.maxSize = maxSize;
        this.occurrences = new HashMap<>();
    }

    public Cache() {
        this(1024);
    }

    public synchronized void setValue(Key key, Value value) {
        if (!data.containsKey(key)) {
            if (data.size() >= maxSize) {
                while (occurrences.get(occurrenceOrder.peek()) > 1) {
                    occurencesTrackedPoll();
                }
                data.remove(occurrenceOrder.poll());
            }
            occurrencesTrackedAdd(key);
        }
        data.put(key, value);
    }


    private void occurrencesTrackedAdd(Key key) {
        int occ = 0;
        if (occurrences.get(key) != null) {
            occ = occurrences.get(key);
        }
        occ++;
        occurrences.put(key, occ);
        occurrenceOrder.add(key);
    }

    private Key occurencesTrackedPoll() {
        Key key = occurrenceOrder.poll();
        int occ = 0;
        if (occurrences.get(key) != null) {
            occ = occurrences.get(key);
        }
        occ--;
        if (occ < 0) throw new IllegalStateException();
        occurrences.put(key, occ);
        return key;
    }

    /**
     * Retrieves cached value associated to the key.
     * Uses equals.
     * If does not exist, returns null.
     * @param key the key associated with the value.
     * @return the value associated to the key.
     */
    public Value getValue(Key key) {
        occurrencesTrackedAdd(key);
        return data.get(key);
    }

    public void clearCache() {
        data.clear();
        occurrenceOrder.clear();
        occurrences.clear();
    }
}